from autoarray.structures import arrays, grids

from typing import Optional

#
# class PointSourceDataset:
#
#     def __init__(self, name, positions : grids.Grid2DIrregular = None, fluxes : arrays.ValuesIrregular):
